from typing import Union
from pydantic import Extra

from unify.agent import Agent
from unify.chat.clients import _Client
from unify.types import Prompt, Score, _FormattedBaseModel, ChatCompletion


class Evaluation(_FormattedBaseModel, extra=Extra.allow, arbitrary_types_allowed=True):
    prompt: Prompt
    response: ChatCompletion
    agent: Union[str, _Client, Agent]
    score: Score



import abc
from typing import Union

import unify
from unify.types import Prompt


class Agent(abc.ABC):

    def __init__(self):
        self._call = unify.with_logging(self._call, endpoint=self.__class__.__name__)

    @abc.abstractmethod
    def _call(self, prompt: Union[str, Prompt]):
        raise NotImplemented

    @abc.abstractmethod
    def __call__(self, prompt: Union[str, Prompt]):
        return self._call(**prompt.dict())



import copy
from typing import Union, List, Dict, Type
from unify.types import _FormattedBaseModel, ChatCompletion, Choice,\
    ChatCompletionMessage

_REPR_MODE = None
_KEYS_TO_SKIP: Dict[Type, Dict] = dict()
_KEYS_TO_KEEP: Dict[Type, Dict] = dict()
_KEYS_TO_KEEP_MODES: Dict[str, Dict[Type, Dict]] = {
    "concise": {
        ChatCompletion: {
            "choices": None
        }
    },
    "verbose": {}
}
_KEYS_TO_SKIP_MODES: Dict[str, Dict[Type, Dict]] = {
    "concise": {
        Choice: {
            "finish_reason": "stop",
            "index": 0
        },
        ChatCompletionMessage: {
            "role": "assistant"
        },
        "verbose": {}
    }
}


def repr_mode() -> str:
    """
    Gets the global representation mode as currently set. This is used when representing
    the various unify types on screen. Can be either "verbose" or "concise".
    """
    global _REPR_MODE
    # noinspection PyProtectedMember
    return _REPR_MODE._val if _REPR_MODE is not None else "verbose"


def key_repr(instance: _FormattedBaseModel) -> Union[Dict, List]:
    """
    Get the key representation for the instance passed, either as the keys to keep or
    the keys to remove.

    Args:
        instance: The instance for which we want to retrieve the key repr policy.

    Returns:
        Dict containing the policy, with a base key of "skip" or "keep", followed by
        the nested structure of the elements to either remove or keep.
    """
    ret = dict()
    fields = instance.model_fields
    if instance.model_extra is not None:
        fields = {**fields, **instance.model_extra}
    for key, field in fields.items():
        annotation = field.annotation
        if annotation in _KEYS_TO_SKIP or annotation in _KEYS_TO_KEEP:
            ret[key] = key_repr(getattr(instance, key))
    ins_type = type(instance)
    to_skip = ins_type in _KEYS_TO_SKIP
    to_keep = ins_type in _KEYS_TO_KEEP
    assert not (to_skip and to_keep),\
        "Cannot have specification for keys to skip AND to keep," \
        "please only set one of these."
    if to_skip:
        return {**ret, **{"skip": _KEYS_TO_SKIP[ins_type]}}
    elif to_keep:
        return {**ret, **{"keep": _KEYS_TO_KEEP[ins_type]}}
    return ret


def keys_to_skip() -> Dict[Type, Dict]:
    """
    Return the currently set keys to skip, which is a dict with types as keys and the
    nested structure to skip as values.
    """
    global _KEYS_TO_SKIP
    return copy.deepcopy(_KEYS_TO_SKIP)


def set_keys_to_skip(skip_keys: Union[Dict[Type, Dict], str]) -> None:
    """
    Set the keys to be skipped during representation, which is a dict with types as keys
    and the nested structure to skip as values.

    Args:
        skip_keys: The types as keywords arguments and dictionary representing the
        structure of the keys to skip for that type as values.
    """
    global _KEYS_TO_SKIP
    if isinstance(skip_keys, str):
        global _KEYS_TO_SKIP_MODES
        _KEYS_TO_SKIP = _KEYS_TO_SKIP_MODES[skip_keys]
        return
    _KEYS_TO_SKIP = skip_keys


def keys_to_keep() -> Dict[Type, Dict]:
    """
    Return the currently set keys to keep, which is a dict with types as keys and the
    nested structure to keep as values.
    """
    global _KEYS_TO_KEEP
    return copy.deepcopy(_KEYS_TO_KEEP)


def set_keys_to_keep(keep_keys: Union[Dict[Type, Dict], str]) -> None:
    """
    Set the keys to be kept during representation, which is a dict with types as keys
    and the nested structure to keep as values.

    Args:
        keep_keys: The types as keywords arguments and dictionary representing the
        structure of the keys to keep for that type as values.
    """
    global _KEYS_TO_KEEP
    if isinstance(keep_keys, str):
        global _KEYS_TO_KEEP_MODES
        _KEYS_TO_KEEP = _KEYS_TO_KEEP_MODES[keep_keys]
        return
    _KEYS_TO_KEEP = keep_keys


def set_repr_mode(mode: str) -> None:
    """
    Sets the global representation mode, to be used when representing the various unify
    types on screen. Can be either "verbose" or "concise".

    Args:
        mode: The value to set the mode to, either "verbose" or "concise".
    """
    global _REPR_MODE, _KEYS_TO_SKIP
    _REPR_MODE = ReprMode(mode)
    set_keys_to_skip(mode)
    set_keys_to_keep(mode)


class ReprMode(str):

    def __init__(self, val: str):
        """
        Set a representation mode for a specific context in the code, by using the
        `with` an instantiation of this class.

        Args:
            val: The value of the string, must be either "verbose" or "concise".
        """
        self._check_str(val)
        # noinspection PyProtectedMember

        self._prev_val = "verbose" if _REPR_MODE is None else repr_mode()
        self._val = val

    @staticmethod
    def _check_str(val: str):
        assert val in ("verbose", "concise"),\
            "Expected value to be either 'verbose' or 'concise', " \
            "but found {}".format(val)

    def __enter__(self) -> None:
        set_repr_mode(self._val)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._val = self._prev_val
        self._prev_val = None
        set_repr_mode(self._val)

    def __repr__(self):
        return str(self._val)

    def __str__(self):
        return str(self._val)


class KeepKeys:

    def __init__(self, keep_keys: Union[Dict[Type, Dict], str]):
        """
        Set a the keys to keep for a specific context in the code, by using the
        `with` an instantiation of this class.

        Args:
            keep_keys: The types as keywords arguments and dictionary representing the
            structure of the keys to keep for that type as values.
        """
        self._prev_val = keys_to_keep()
        self._val = keep_keys

    @staticmethod
    def _check_str(val: str):
        assert val in ("verbose", "concise"),\
            "Expected value to be either 'verbose' or 'concise', " \
            "but found {}".format(val)

    def __enter__(self) -> None:
        set_keys_to_keep(self._val)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._val = self._prev_val
        self._prev_val = None
        set_keys_to_keep(self._val)

    def __repr__(self):
        return str(self._val)

    def __str__(self):
        return str(self._val)


class SkipKeys:

    def __init__(self, skip_keys: Union[Dict[Type, Dict], str]):
        """
        Set a the keys to skip for a specific context in the code, by using the
        `with` an instantiation of this class.

        Args:
            skip_keys: The types as keywords arguments and dictionary representing the
            structure of the keys to skip for that type as values.
        """
        self._prev_val = keys_to_keep()
        self._val = skip_keys

    @staticmethod
    def _check_str(val: str):
        assert val in ("verbose", "concise"), \
            "Expected value to be either 'verbose' or 'concise', " \
            "but found {}".format(val)

    def __enter__(self) -> None:
        set_keys_to_skip(self._val)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._val = self._prev_val
        self._prev_val = None
        set_keys_to_skip(self._val)

    def __repr__(self):
        return str(self._val)

    def __str__(self):
        return str(self._val)


# noinspection PyRedeclaration
_REPR_MODE = ReprMode("verbose")



"""Unify python module."""
import os

if "UNIFY_BASE_URL" in os.environ.keys():
    BASE_URL = os.environ["UNIFY_BASE_URL"]
else:
    BASE_URL = "https://api.unify.ai/v0"

from .utils import (
    credits,
    custom_api_keys,
    custom_endpoints,
    datasets,
    efficiency_benchmarks,
    evaluations,
    evaluators,
    helpers,
    logging,
    router_configurations,
    router_deployment,
    router_training,
    supported_endpoints,
)
from .utils.credits import *
from .utils.custom_api_keys import *
from .utils.custom_endpoints import *
from .utils.datasets import *
from .utils.efficiency_benchmarks import *
from .utils.evaluations import *
from .utils.evaluators import *
from .utils.helpers import *
from .utils.logging import *
from .utils.router_configurations import *
from .utils.router_deployment import *
from .utils.router_training import *
from .utils.supported_endpoints import *

from .chat import chatbot, clients
from .chat.clients import multi_llm
from .chat.chatbot import *
from unify.chat.clients.uni_llm import *
from unify.chat.clients.multi_llm import *

from .agent import *
from .casting import *
from .dataset import *
from .evaluation import *
from .evaluator import *
from .logging import *
from .repr import *
from .types import *



from __future__ import annotations
from typing_extensions import Self
from typing import List, Dict, Union, Optional

import unify
from unify.casting import cast
from unify.types import _Formatted
from unify.types import Prompt, Datum
# noinspection PyProtectedMember
from .utils.helpers import _validate_api_key, _dict_aligns_with_pydantic


class Dataset(_Formatted):
    def __init__(
        self,
        data: Union[str, Dict, Prompt, Datum,
                    List[Union[str, Dict, Prompt, Datum]]],
        *,
        name: str = None,
        auto_sync: Union[bool, str] = False,
        api_key: Optional[str] = None,
    ) -> None:
        """
        Initialize a local dataset of LLM queries.

        Args:
            data: The data for populating the dataset. This can either can a list of
            user messages, a list of full queries, or a list of dicts of queries
            alongside any extra fields. Individual items in any of the formats listed
            above will also be converted to lists and processed automatically.

            name: The name of the dataset.

            auto_sync: Whether to automatically keep this dataset fully synchronized
            with the upstream variant at all times. If `True` or "both" then the sync
            will be bi-directional, if "upload" then all local changes will be
            uploaded to the upstream account without any downloads, if "download"
            then all upstream changes will be downloaded locally without any uploads.
            If "upstream_mirrors_local" then the upstream dataset will be anchored to
            the local version at all times, and any other uploads outside the local
            dataset will be overwritten. If "local_mirrors_upstream" then the local
            version will be anchored to the upstream version at all times, and any local
            changes will be overwritten. If `False` or "neither" then no synchronization
            will be done automatically.

            api_key: API key for accessing the Unify API. If None, it attempts to
            retrieve the API key from the environment variable UNIFY_KEY. Defaults to
            None.

        Raises:
            UnifyError: If the API key is missing.
        """
        self._name = name
        if isinstance(data, tuple):
            data = list(data)
        elif not isinstance(data, list):
            data = [data]
        if isinstance(data[0], str):
            self._data = [cast(usr_msg, Datum) for usr_msg in data]
        elif isinstance(data[0], Prompt):
            self._data = [cast(prompt, Datum) for prompt in data]
        elif isinstance(data[0], dict) and _dict_aligns_with_pydantic(data[0], Prompt):
            self._data = [cast(Prompt(**dct), Datum) for dct in data]
        elif isinstance(data[0], Datum):
            self._data = data
        elif isinstance(data[0], dict) and \
                _dict_aligns_with_pydantic(data[0], Datum):
            self._data = self._data = [Datum(**dct) for dct in data]
        else:
            raise Exception("input {} with entries of type {} does not align with "
                            "expected input types.".format(data, type(data[0])))
        self._api_key = _validate_api_key(api_key)
        self._auto_sync_flag = auto_sync
        self._auto_sync()

    @property
    def name(self) -> str:
        """
        Name of the dataset.
        """
        self._auto_sync()
        return self._name

    @property
    def auto_sync(self) -> Union[bool, str]:
        """
        The auto-sync mode currently selected.
        This dictates whether to automatically keep this dataset fully synchronized
        with the upstream variant at all times. If `True` or "both" then the sync
        will be bi-directional, if "upload" then all local changes will be
        uploaded to the upstream account without any downloads, if "download"
        then all upstream changes will be downloaded locally without any uploads.
        If "upstream_mirrors_local" then the upstream dataset will be anchored to the
        local version at all times, and any other uploads outside the local dataset
        will be overwritten. If "local_mirrors_upstream" then the local version will be
        anchored to the upstream version at all times, and any local changes will be
        overwritten. If `False` or "neither" then no synchronization will be done
        automatically.
        """
        self._auto_sync()
        return self._auto_sync_flag

    def set_name(self, name: str) -> Self:
        """
        Set the name of the dataset.

        Args:
            name: The name to set the dataset to.

        Returns:
            This dataset, useful for chaining methods.
        """
        self._auto_sync()
        self._name = name
        return self

    def set_auto_sync(self, auto_sync: Union[bool, str]) -> Self:
        """
        Set the value of the auto-sync flag.

        Args:
            auto_sync: Whether to automatically keep this dataset fully synchronized
            with the upstream variant at all times. If `True` or "both" then the sync
            will be bi-directional, if "upload" then all local changes will be
            uploaded to the upstream account without any downloads, if "download"
            then all upstream changes will be downloaded locally without any uploads.
            If "upstream_mirrors_local" then the upstream dataset will be anchored to
            the local version at all times, and any other uploads outside the local
            dataset will be overwritten. If "local_mirrors_upstream" then the local
            version will be anchored to the upstream version at all times, and any local
            changes will be overwritten. If `False` or "neither" then no synchronization
            will be done automatically.

        Returns:
            This dataset, useful for chaining methods.
        """
        self._auto_sync()
        self._auto_sync_flag = auto_sync
        return self

    @staticmethod
    def from_upstream(
        name: str,
        auto_sync: Union[bool, str] = False,
        api_key: Optional[str] = None,
    ) -> Self:
        """
        Initialize a local dataset of LLM queries, from the upstream dataset.

        Args:
            name: The name of the dataset.

            auto_sync: Whether to automatically keep this dataset fully synchronized
            with the upstream variant at all times.

            api_key: API key for accessing the Unify API. If None, it attempts to
            retrieve the API key from the environment variable UNIFY_KEY. Defaults to
            None.

        Returns:
            The dataset, with contents downloaded from upstream.

        Raises:
            UnifyError: If the API key is missing.
        """
        data = unify.download_dataset(name, api_key=api_key)
        return Dataset(data, name=name, auto_sync=auto_sync, api_key=api_key)

    def _assert_name_exists(self) -> None:
        assert self._name is not None, (
            "Dataset name must be specified in order to upload, download, sync or "
            "compare to a corresponding dataset in your upstream account. "
            "You can simply use .set_name() and set it to the same name as your "
            "upstream dataset, or create a new name if it doesn't yet exist upstream."
        )

    def upload(self, overwrite: bool = False) -> Self:
        """
        Uploads all unique local data in the dataset to the user account upstream.
        This function will not download any uniques from upstream.
        Use `sync` to synchronize and superset the datasets in both directions.
        Set `overwrite=True` to disregard any pre-existing upstream data.

        Args:
            overwrite: Whether to overwrite the upstream dataset if it already exists.

        Returns:
            This dataset, useful for chaining methods.
        """
        if self._auto_sync_flag == "local_mirrors_upstream":
            raise Exception("upload not permitted when local mirrors upstream")
        self._assert_name_exists()
        dataset_exists_upstream = self._name in unify.list_datasets(self._api_key)
        if overwrite:
            if dataset_exists_upstream:
                upstream_dataset = unify.download_dataset(
                    self._name, api_key=self._api_key
                )
                unique_upstream = [item.dict() for item in upstream_dataset
                                   if item not in self._data]
                unify.delete_data(self._name, unique_upstream)
                unique_local = [item.dict() for item in self._data
                                if item not in upstream_dataset]
                unify.add_data(self._name, unique_local)
            else:
                data = [datum.dict() for datum in self._data]
                unify.upload_dataset_from_dictionary(self._name, data)
        else:
            if dataset_exists_upstream:
                upstream_dataset = unify.download_dataset(
                    self._name, api_key=self._api_key
                )
                unique_local = [item.dict() for item in self._data
                                if item not in upstream_dataset]
                unify.add_data(self._name, unique_local)
            else:
                unique_local = [entry.dict() for entry in self._data]
                unify.upload_dataset_from_dictionary(self._name, unique_local)
        if self._auto_sync_flag in (True, "both", "download", "local_mirrors_upstream"):
            auto_sync_flag = self._auto_sync_flag
            self._auto_sync_flag = False
            overwrite = True if auto_sync_flag == "local_mirrors_upstream" else False
            self.download(overwrite=overwrite)
            self._auto_sync_flag = auto_sync_flag
        return self

    def download(self, overwrite: bool = False) -> Self:
        """
        Downloads all unique upstream data from the user account to the local dataset.
        This function will not upload any unique values stored locally.
        Use `sync` to synchronize and superset the datasets in both directions.
        Set `overwrite=True` to disregard any pre-existing data stored in this class.

        Args:
            overwrite: Whether to overwrite the local data, if any already exists

        Returns:
            This dataset after the in-place download, useful for chaining methods.
        """
        if self._auto_sync_flag == "upstream_mirrors_local":
            raise Exception("download not permitted when upstream mirrors local")
        self._assert_name_exists()
        if overwrite:
            self._data = unify.download_dataset(self._name, api_key=self._api_key)
        else:
            upstream_dataset = unify.download_dataset(
                self._name, api_key=self._api_key
            )
            unique_local = [item for item in self._data if item not in upstream_dataset]
            self._data = upstream_dataset + unique_local
        if self._auto_sync_flag in (True, "both", "upload", "upstream_mirrors_local"):
            auto_sync_flag = self._auto_sync_flag
            self._auto_sync_flag = False
            overwrite = True if auto_sync_flag == "upstream_mirrors_local" else False
            self.upload(overwrite=overwrite)
            self._auto_sync_flag = auto_sync_flag
        return self

    def _auto_sync(self) -> None:
        if self._auto_sync_flag in (True, "both", "download", "local_mirrors_upstream"):
            auto_sync_flag = self._auto_sync_flag
            self._auto_sync_flag = False
            overwrite = True if auto_sync_flag == "local_mirrors_upstream" else False
            self.download(overwrite=overwrite)
            self._auto_sync_flag = auto_sync_flag
        if self._auto_sync_flag in (True, "both", "upload", "upstream_mirrors_local"):
            auto_sync_flag = self._auto_sync_flag
            self._auto_sync_flag = False
            overwrite = True if auto_sync_flag == "upstream_mirrors_local" else False
            self.upload(overwrite=overwrite)
            self._auto_sync_flag = auto_sync_flag

    def sync(self) -> Self:
        """
        Synchronize the dataset in both directions, downloading any values missing
        locally, and uploading any values missing from upstream in the account.

        Returns:
            This dataset after the in-place sync, useful for chaining methods.
        """
        auto_sync_flag = self._auto_sync_flag
        self._auto_sync_flag = False
        self.download()
        self.upload()
        self._auto_sync_flag = auto_sync_flag
        return self

    def upstream_diff(self) -> Self:
        """
        Prints the difference between the local dataset and the upstream dataset.

        Returns:
            This dataset after printing the diff, useful for chaining methods.
        """
        self._assert_name_exists()
        upstream_dataset = unify.download_dataset(self._name, api_key=self._api_key)
        unique_upstream = [item for item in upstream_dataset if item not in self._data]
        print(
            "The following {} queries are stored upstream but not locally\n: "
            "{}".format(len(unique_upstream), unique_upstream)
        )
        unique_local = [item for item in self._data if item not in upstream_dataset]
        print(
            "The following {} queries are stored upstream but not locally\n: "
            "{}".format(len(unique_local), unique_local)
        )
        self._auto_sync()
        return self

    def add(self, other: Union[Dataset, str, Dict, Prompt, Datum,
                               List[Union[str, Dict, Prompt, Datum]]]) -> Self:
        """
        Adds another dataset to this one, return a new Dataset instance, with this
        new dataset receiving all unique queries from the other added dataset.

        Args:
            other: The other dataset being added to this one.

        Returns:
            The new dataset following the addition.
        """
        other = other if isinstance(other, Dataset) else Dataset(other)
        data = list(dict.fromkeys(self._data + other._data))
        return Dataset(data=data, auto_sync=self._auto_sync_flag, api_key=self._api_key)

    def sub(self, other: Union[Dataset, str, Dict, Prompt, Datum,
                               List[Union[str, Dict, Prompt, Datum]]]) -> Self:
        """
        Subtracts another dataset from this one, return a new Dataset instance, with
        this new dataset losing all queries from the other subtracted dataset.

        Args:
            other: The other dataset being added to this one.

        Returns:
            The new dataset following the subtraction.
        """
        other = other if isinstance(other, Dataset) else Dataset(other)
        assert other in self, (
            "cannot subtract dataset B from dataset A unless all queries of dataset "
            "B are also present in dataset A"
        )
        data = [item for item in self._data if item not in other]
        return Dataset(data=data, auto_sync=self._auto_sync_flag, api_key=self._api_key)

    def inplace_add(
            self,
            other: Union[Dataset, str, Dict, Prompt, Datum,
                         List[Union[str, Dict, Prompt, Datum]]]
    ) -> Self:
        """
        Adds another dataset to this one, with this dataset receiving all unique queries
        from the other added dataset.

        Args:
            other: The other dataset being added to this one.

        Returns:
            This dataset following the in-place addition.
        """
        if self._auto_sync_flag == "local_mirrors_upstream":
            raise TypeError("Adding entries not permitted when local mirrors upstream")
        other = other if isinstance(other, Dataset) else Dataset(other)
        self._data = list(dict.fromkeys(self._data + other._data))
        self._auto_sync()
        return self

    def inplace_sub(
            self,
            other: Union[Dataset, str, Dict, Prompt, Datum,
                         List[Union[str, Dict, Prompt, Datum]]]
    ) -> Self:
        """
        Subtracts another dataset from this one, with this dataset losing all queries
        from the other subtracted dataset.

        Args:
            other: The other dataset being added to this one.

        Returns:
            This dataset following the in-place subtraction.
        """
        if self._auto_sync_flag == "local_mirrors_upstream":
            raise TypeError("Adding entries not permitted when local mirrors upstream")
        other = other if isinstance(other, Dataset) else Dataset(other)
        assert other in self, (
            "cannot subtract dataset B from dataset A unless all queries of dataset "
            "B are also present in dataset A"
        )
        self._data = [item for item in self._data if item not in other]
        self._auto_sync()
        return self

    def __add__(self, other: Union[Dataset, str, Dict, Prompt, Datum,
                                   List[Union[str, Dict, Prompt, Datum]]]) -> Self:
        """
        Adds another dataset to this one via the + operator, return a new Dataset
        instance, with this new dataset receiving all unique queries from the other
        added dataset.

        Args:
            other: The other dataset being added to this one.

        Returns:
            The new dataset following the addition.
        """
        return self.add(other)

    def __radd__(self, other: Union[Dataset, str, Dict, Prompt, Datum,
                                    List[Union[str, Dict, Prompt, Datum]]]) -> Self:
        """
        Adds another dataset to this one via the + operator, this is used if the
        other item does not have a valid __add__ method for these two types. Return a
        new Dataset instance, with this new dataset receiving all unique queries from
        the other added dataset.

        Args:
            other: The other dataset being added to this one.

        Returns:
            The new dataset following the addition.
        """
        return Dataset(other).add(self)

    def __iadd__(self, other: Union[Dataset, str, Dict, Prompt, Datum,
                                    List[Union[str, Dict, Prompt, Datum]]]) -> Self:
        """
        Adds another dataset to this one, with this dataset receiving all unique queries
        from the other added dataset.

        Args:
            other: The other dataset being added to this one.

        Returns:
            This dataset following the in-place addition.
        """
        return self.inplace_add(other)

    def __sub__(self, other: Union[Dataset, str, Dict, Prompt, Datum,
                                   List[Union[str, Dict, Prompt, Datum]]]) -> Self:
        """
        Subtracts another dataset from this one via the - operator, return a new Dataset
        instance, with this new dataset losing all queries from the other subtracted
        dataset.

        Args:
            other: The other dataset being subtracted from this one.

        Returns:
            The new dataset following the subtraction.
        """
        return self.sub(other)

    def __rsub__(self, other: Union[Dataset, str, Dict, Prompt, Datum,
                                    List[Union[str, Dict, Prompt, Datum]]]) -> Self:
        """
        Subtracts another dataset from this one via the - operator, this is used if the
        other item does not have a valid __sub__ method for these two types. Return a
        new Dataset instance, with this new dataset losing all queries from the other
        subtracted dataset.

        Args:
            other: The other dataset being subtracted from this one.

        Returns:
            The new dataset following the subtraction.
        """
        return Dataset(other).sub(self)

    def __isub__(
            self,
            other: Union[Dataset, str, Dict, Prompt, Datum,
                         List[Union[str, Dict, Prompt, Datum]]]
    ) -> Self:
        """
        Subtracts another dataset from this one, with this dataset losing all queries
        from the other subtracted dataset.

        Args:
            other: The other dataset being added to this one.

        Returns:
            This dataset following the in-place subtraction.
        """
        return self.inplace_sub(other)

    def __iter__(self) -> Datum:
        """
        Iterates through the dataset, return one Datum instance at a time.

        Returns:
            A Datum instance per iteration.
        """
        self._auto_sync()
        for x in self._data:
            yield x

    def __contains__(self, item: Union[Dataset, str, Dict, Prompt, Datum,
                                       List[Union[str, Dict, Prompt, Datum]]]) -> bool:
        """
        Determine whether the item is contained within the dataset. The item is cast to
        a Dataset instance, and can therefore take on many different types. Only returns
        True if *all* entries in the passed dataset are contained within this dataset.

        Args:
            item: The item to cast to a Dataset before checking if it's a subset of this
            one.

        Returns:
            Boolean, as to whether or not the passed Dataset is a subset of this one.
        """
        item = item if isinstance(item, Dataset) else Dataset(item)
        this_set = set(self._data)
        combined_set = set(self._data + item._data)
        return len(this_set) == len(combined_set)

    def __len__(self) -> int:
        """
        Returns the number of entries contained within the dataset.

        Returns:
            The number of entries in the dataset.
        """
        return len(self._data)

    def __getitem__(self, item: Union[int, slice]) -> Union[Datum, Dataset]:
        """
        Gets an item from the dataset, either via an int or slice. In the case of an
        int then a Datum instance is returned, and for a slice a Dataset instance is
        returned.

        Args:
            item: integer or slice for extraction.

        Returns:
            A Datum or Dataset instance, for int and slice queries respectively.
        """
        self._auto_sync()
        if isinstance(item, int):
            return self._data[item]
        elif isinstance(item, slice):
            return Dataset(self._data[item.start:item.stop:item.step])
        raise TypeError("expected item to be of type int or slice,"
                        "but found {} of type {}".format(item, type(item)))

    def __rich_repr__(self) -> List[Datum]:
        """
        Used by the rich package for representing and print the instance.
        """
        self._auto_sync()
        yield self._data



from typing import Union, Type
from unify.types import Prompt, Datum, ChatCompletion, Score


# Upcasting

def _usr_msg_to_prompt(user_message: str) -> Prompt:
    return Prompt(user_message)


def _prompt_to_datum(prompt: Prompt) -> Datum:
    return Datum(prompt=prompt)


def _usr_msg_to_datum(user_message: str) -> Datum:
    return _prompt_to_datum(_usr_msg_to_prompt(user_message))


def _assis_msg_to_chat_completion(assistant_message: str) -> ChatCompletion:
    return ChatCompletion(assistant_message)


def _bool_to_float(boolean: bool) -> float:
    return float(boolean)


def _float_to_score(float_in: float, score_type: Type[Score]) -> Score:
    return score_type(float_in)


def _bool_to_score(boolean: bool, score_type: Type[Score]) -> Score:
    return _float_to_score(_bool_to_float(boolean), score_type)


# Downcasting

def _prompt_to_usr_msg(prompt: Prompt) -> str:
    return prompt.messages[-1]["content"]


def _datum_to_prompt(datum: Datum) -> Prompt:
    return datum.prompt


def _datum_to_usr_msg(datum: Datum) -> str:
    return _prompt_to_usr_msg(_datum_to_prompt(datum))


def _chat_completion_to_assis_msg(chat_completion: ChatCompletion) -> str:
    return chat_completion.choices[0].message.content


def _float_to_bool(float_in: float) -> bool:
    return bool(float_in)


def _score_to_float(score: Score) -> float:
    return score.score[0]


def _score_to_bool(score: Score) -> bool:
    return _float_to_bool(_score_to_float(score))


# Cast Dict

_CAST_DICT = {
    str: {
        Prompt: _usr_msg_to_prompt,
        Datum: _usr_msg_to_datum,
        ChatCompletion: _assis_msg_to_chat_completion
    },
    Prompt: {
        str: _prompt_to_usr_msg,
        Datum: _prompt_to_datum
    },
    Datum: {
        str: _datum_to_usr_msg,
        Prompt: _datum_to_prompt
    },
    ChatCompletion: {
        str: _chat_completion_to_assis_msg
    },
    bool: {
        float: _bool_to_float,
        Score: _bool_to_score
    },
    float: {
        bool: _float_to_bool,
        Score: _float_to_score
    },
    Score: {
        bool: _score_to_bool,
        float: _score_to_float
    }
}


# Public function

def cast(
        input: Union[str, bool, float, Score, Prompt, Datum, ChatCompletion],
        to_type: Type[Union[str, bool, float, Score, Prompt, Datum, ChatCompletion]],
) -> Union[str, bool, float, Score, Prompt, Datum, ChatCompletion]:
    """
    Cast the input to the specified type.

    Args:
        input: The input to cast.

        to_type: The type to cast the input to.

    Returns:
        The input after casting to the new type.
    """
    input_type = type(input)
    if input_type is to_type:
        return input
    if issubclass(to_type, Score):
        return _CAST_DICT[input_type][Score](input, to_type)
    return _CAST_DICT[input_type][to_type](input)



import os
import json
import requests
from pydantic import BaseModel, ValidationError
from typing import List, Dict, Optional, Union, Any


def _res_to_list(response: requests.Response) -> Union[List, Dict]:
    return json.loads(response.text)


def _validate_api_key(api_key: Optional[str]) -> str:
    if api_key is None:
        api_key = os.environ.get("UNIFY_KEY")
    if api_key is None:
        raise KeyError(
            "UNIFY_KEY is missing. Please make sure it is set correctly!",
        )
    return api_key


def _default(value: Any, default_value: Any) -> Any:
    return value if value is not None else default_value


def _dict_aligns_with_pydantic(dict_in: Dict, pydantic_cls: type(BaseModel)) -> bool:
    try:
        pydantic_cls.model_validate(dict_in)
        return True
    except ValidationError:
        return False



import os
import requests
from typing import Optional, Any, Dict

from unify import BASE_URL
from .helpers import _validate_api_key


def trigger_evaluation(
    evaluator: str,
    dataset: str,
    endpoint: str,
    client_side_scores: Optional[str] = None,
    api_key: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Trigger an evaluation for a specific dataset using a given evaluator and endpoint.

    Args:
        evaluator: Name of the evaluator to use.
        dataset: Name of the uploaded dataset to evaluate.
        endpoint: Name of the endpoint to evaluate. Must be specified using the `model@provider` format.
        client_side_scores: Optional path to a JSONL file containing client-side scores.
        api_key: If specified, unify API key to be used. Defaults to the value in the `UNIFY_KEY` environment variable.

    Returns:
        A dictionary containing the response from the API.

    Raises:
        requests.HTTPError: If the API request fails.
        KeyError: If the API key is not provided and not set in the environment.
        FileNotFoundError: If the client_side_scores file is specified but not found.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    url = f"{BASE_URL}/evaluation"

    params = {
        "evaluator": evaluator,
        "dataset": dataset,
        "endpoint": endpoint,
    }

    files = {}
    if client_side_scores:
        if not os.path.exists(client_side_scores):
            raise FileNotFoundError(
                f"Client-side scores file not found: {client_side_scores}"
            )
        files["client_side_scores"] = (
            "client_scores.jsonl",
            open(client_side_scores, "rb"),
            "application/json",
        )

    response = requests.post(url, headers=headers, params=params, files=files)
    response.raise_for_status()

    return response.json()


def get_evaluations(
    dataset: str,
    endpoint: Optional[str] = None,
    evaluator: Optional[str] = None,
    per_prompt: bool = False,
    api_key: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Get evaluations for a specific dataset, optionally filtered by endpoint and evaluator.

    Args:
        dataset: Name of the dataset to fetch evaluation from.
        endpoint: The endpoint to fetch the evaluation for. If None, returns evaluations for all endpoints.
        evaluator: Name of the evaluator to fetch the evaluation for. If None, returns all available evaluations for the dataset and endpoint pair.
        per_prompt: If True, returns the scores on a per-prompt level. By default set to False. If True requires an eval name and endpoint to be set.
        api_key: If specified, unify API key to be used. Defaults to the value in the `UNIFY_KEY` environment variable.

    Returns:
        A dictionary containing the evaluation results.

    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }

    url = f"{BASE_URL}/evaluation"

    params = {"dataset": dataset, "per_prompt": per_prompt}

    if endpoint:
        params["endpoint"] = endpoint
    if evaluator:
        params["evaluator"] = evaluator

    response = requests.get(url, headers=headers, params=params)
    response.raise_for_status()

    return response.json()


def delete_evaluations(
    dataset: str,
    endpoint: Optional[str] = None,
    evaluator: Optional[str] = None,
    api_key: Optional[str] = None,
) -> Dict[str, str]:
    """
    Delete evaluations for a specific dataset, optionally filtered by endpoint and evaluator.

    Args:
        dataset: Name of the dataset to delete the evaluation for.
        endpoint: The endpoint to delete the evaluation for. If None, deletes the evaluations for all endpoints.
        evaluator: Name of the evaluator to delete the evaluation for. If None, deletes all available evaluations for the dataset and endpoint pair.
        api_key: If specified, unify API key to be used. Defaults to the value in the `UNIFY_KEY` environment variable.

    Returns:
        A dictionary containing information about the deletion operation.

    Raises:
        requests.exceptions.RequestException: If the API request fails.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }

    url = f"{BASE_URL}/evaluation"

    params = {
        "dataset": dataset,
    }

    if endpoint:
        params["endpoint"] = endpoint
    if evaluator:
        params["evaluator"] = evaluator

    response = requests.delete(url, headers=headers, params=params)
    response.raise_for_status()

    return response.json()



def train_router():
    raise NotImplementedError


def delete_router():
    raise NotImplementedError


def rename_router():
    raise NotImplementedError


def list_routers():
    raise NotImplementedError



import os
import json
from typing import Dict, Union, Optional
from unify.types import ChatCompletion

_cache: Optional[Dict] = None
_cache_dir = os.environ["UNIFY_CACHE_DIR"] if "UNIFY_CACHE_DIR" in os.environ \
    else os.getcwd()
_cache_fpath: str = os.path.join(_cache_dir, ".cache.json")


def _create_cache_if_none():
    global _cache
    if _cache is None:
        if not os.path.exists(_cache_fpath):
            with open(_cache_fpath, "w") as outfile:
                json.dump({}, outfile)
        with open(_cache_fpath) as outfile:
            _cache = json.load(outfile)


# noinspection PyTypeChecker,PyUnboundLocalVariable
def _get_cache(kw: Dict) -> Union[None, Dict]:
    _create_cache_if_none()
    kw = {k: v for k, v in kw.items() if v is not None}
    kw_str = json.dumps(kw)
    if kw_str in _cache:
        return ChatCompletion(**json.loads(_cache[kw_str]))


# noinspection PyTypeChecker,PyUnresolvedReferences
def _write_to_cache(kw, response):
    _create_cache_if_none()
    kw = {k: v for k, v in kw.items() if v is not None}
    kw_str = json.dumps(kw)
    response_str = json.dumps(response.model_dump())
    _cache[kw_str] = response_str
    with open(_cache_fpath, "w") as outfile:
        json.dump(_cache, outfile)



import requests
from typing import Optional, List, Any, Dict

from unify import BASE_URL
from .helpers import _validate_api_key


def create_evaluator(
    evaluator_config: Dict[str, Any], api_key: Optional[str] = None
) -> Dict[str, str]:
    """
    Create a new evaluator based on the provided configuration.

    Args:
        evaluator_config: A dictionary containing the configuration for the evaluator.
        api_key: If specified, unify API key to be used. Defaults
        to the value in the `UNIFY_KEY` environment variable.

    Returns:
        A dictionary containing information about the created evaluator.

    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Content-Type": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    url = f"{BASE_URL}/evaluator"

    response = requests.post(url, headers=headers, json=evaluator_config)
    response.raise_for_status()

    return response.json()


def get_evaluator(name: str, api_key: Optional[str] = None) -> Dict[str, Any]:
    """
    Get the configuration of a specific evaluator.

    Args:
        name: Name of the evaluator to return the configuration of.
        api_key: If specified, unify API key to be used. Defaults
        to the value in the `UNIFY_KEY` environment variable.

    Returns:
        A dictionary containing the evaluator configuration if successful.

    Raises:
        requests.HTTPError: If the API request fails.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    url = f"{BASE_URL}/evaluator"

    params = {"name": name}

    response = requests.get(url, headers=headers, params=params)
    response.raise_for_status()

    return response.json()


def delete_evaluator(name: str, api_key: Optional[str] = None) -> Dict[str, str]:
    """
    Delete an evaluator by its name.

    Args:
        name: Name of the evaluator to delete.
        api_key: If specified, unify API key to be used. Defaults
        to the value in the `UNIFY_KEY` environment variable.

    Returns:
        A dictionary containing the response message if successful.

    Raises:
        requests.HTTPError: If the API request fails.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    url = f"{BASE_URL}/evaluator"

    params = {"name": name}

    response = requests.delete(url, headers=headers, params=params)
    response.raise_for_status()

    return response.json()


def rename_evaluator(
    name: str, new_name: str, api_key: Optional[str] = None
) -> Dict[str, str]:
    """
    Rename an existing evaluator.

    Args:
        name: Name of the evaluator to rename.
        new_name: New name for the evaluator.
        api_key: If specified, unify API key to be used. Defaults
        to the value in the `UNIFY_KEY` environment variable.

    Returns:
        A dictionary containing the response message if successful.

    Raises:
        requests.HTTPError: If the API request fails.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    url = f"{BASE_URL}/evaluator/rename"

    params = {"name": name, "new_name": new_name}

    response = requests.post(url, headers=headers, params=params)
    response.raise_for_status()

    return response.json()


def list_evaluators(api_key: Optional[str] = None) -> List[str]:
    """
    Get a list of available evaluators.

    Args:
        api_key: If specified, unify API key to be used. Defaults
        to the value in the `UNIFY_KEY` environment variable.

    Returns:
        A list of available evaluator names if successful, otherwise an empty list.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    url = f"{BASE_URL}/evaluator/list"
    response = requests.get(url, headers=headers)
    response.raise_for_status()

    return response.json()



# global
import requests
# noinspection PyProtectedMember
from openai._types import Headers, Query, Body
from openai.types.chat import (
    ChatCompletionToolParam,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionMessageParam,
    ChatCompletionStreamOptionsParam,
)
from openai.types.chat.completion_create_params import ResponseFormat
from abc import ABC, abstractmethod
from typing import Mapping, Dict, List, Optional, Union, Iterable
from typing_extensions import Self

# local
from unify import BASE_URL
from unify.types import Prompt
# noinspection PyProtectedMember
from unify.utils.helpers import _validate_api_key, _default


class _Client(ABC):
    """Base Abstract class for interacting with the Unify chat completions endpoint."""

    def __init__(
            self,
            *,
            system_message: Optional[str],
            messages: Optional[
                Union[List[ChatCompletionMessageParam],
                      Dict[str, List[ChatCompletionMessageParam]]]],
            frequency_penalty: Optional[float],
            logit_bias: Optional[Dict[str, int]],
            logprobs: Optional[bool],
            top_logprobs: Optional[int],
            max_completion_tokens: Optional[int],
            n: Optional[int],
            presence_penalty: Optional[float],
            response_format: Optional[ResponseFormat],
            seed: Optional[int],
            stop: Union[Optional[str], List[str]],
            stream: Optional[bool],
            stream_options: Optional[ChatCompletionStreamOptionsParam],
            temperature: Optional[float],
            top_p: Optional[float],
            tools: Optional[Iterable[ChatCompletionToolParam]],
            tool_choice: Optional[ChatCompletionToolChoiceOptionParam],
            parallel_tool_calls: Optional[bool],
            # platform arguments
            use_custom_keys: bool,
            tags: Optional[List[str]],
            drop_params: Optional[bool],
            region: Optional[str] = None,
            log_query_body: Optional[bool],
            log_response_body: Optional[bool],
            api_key: Optional[str],
            # python client arguments
            return_full_completion: bool,
            cache: bool,
            # passthrough arguments
            extra_headers: Optional[Headers],
            extra_query: Optional[Query],
            **kwargs,
    ) -> None:  # noqa: DAR101, DAR401
        self._api_key = _validate_api_key(api_key)
        self._system_message = None
        self.set_system_message(system_message)
        self._messages = None
        self.set_messages(messages)
        self._frequency_penalty = None
        self.set_frequency_penalty(frequency_penalty)
        self._logit_bias = None
        self.set_logit_bias(logit_bias)
        self._logprobs = None
        self.set_logprobs(logprobs)
        self._top_logprobs = None
        self.set_top_logprobs(top_logprobs)
        self._max_completion_tokens = None
        self.set_max_completion_tokens(max_completion_tokens)
        self._n = None
        self.set_n(n)
        self._presence_penalty = None
        self.set_presence_penalty(presence_penalty)
        self._response_format = None
        self.set_response_format(response_format)
        self._seed = None
        self.set_seed(seed)
        self._stop = None
        self.set_stop(stop)
        self._stream = None
        self.set_stream(stream)
        self._stream_options = None
        self.set_stream_options(stream_options)
        self._temperature = None
        self.set_temperature(temperature)
        self._top_p = None
        self.set_top_p(top_p)
        self._tools = None
        self.set_tools(tools)
        self._tool_choice = None
        self.set_tool_choice(tool_choice)
        self._parallel_tool_calls = None
        self.set_parallel_tool_calls(parallel_tool_calls)
        # platform arguments
        self._use_custom_keys = None
        self.set_use_custom_keys(use_custom_keys)
        self._tags = None
        self.set_tags(tags)
        self._drop_params = None
        self.set_drop_params(drop_params)
        self._region = None
        self.set_region(region)
        self._log_query_body = None
        self.set_log_query_body(log_query_body)
        self._log_response_body = None
        self.set_log_response_body(log_response_body)
        # python client arguments
        self._return_full_completion = None
        self.set_return_full_completion(return_full_completion)
        self._cache = None
        self.set_cache(cache)
        # passthrough arguments
        self._extra_headers = None
        self.set_extra_headers(extra_headers)
        self._extra_query = None
        self.set_extra_query(extra_query)
        self._extra_body = None
        self.set_extra_body(kwargs)

    # Properties #
    # -----------#

    @property
    def system_message(self) -> Optional[str]:
        """
        Get the default system message, if set.

        Returns:
            The default system message.
        """
        return self._system_message

    @property
    def messages(self) -> Optional[
        Union[List[ChatCompletionMessageParam],
              Dict[str, List[ChatCompletionMessageParam]]]]:
        """
        Get the default messages, if set.

        Returns:
            The default messages.
        """
        return self._messages

    @property
    def frequency_penalty(self) -> Optional[float]:
        """
        Get the default frequency penalty, if set.

        Returns:
            The default frequency penalty.
        """
        return self._frequency_penalty

    @property
    def logit_bias(self) -> Optional[Dict[str, int]]:
        """
        Get the default logit bias, if set.

        Returns:
            The default logit bias.
        """
        return self._logit_bias

    @property
    def logprobs(self) -> Optional[bool]:
        """
        Get the default logprobs, if set.

        Returns:
            The default logprobs.
        """
        return self._logprobs

    @property
    def top_logprobs(self) -> Optional[int]:
        """
        Get the default top logprobs, if set.

        Returns:
            The default top logprobs.
        """
        return self._top_logprobs

    @property
    def max_completion_tokens(self) -> Optional[int]:
        """
        Get the default max tokens, if set.

        Returns:
            The default max tokens.
        """
        return self._max_completion_tokens

    @property
    def n(self) -> Optional[int]:
        """
        Get the default n, if set.

        Returns:
            The default n value.
        """
        return self._n

    @property
    def presence_penalty(self) -> Optional[float]:
        """
        Get the default presence penalty, if set.

        Returns:
            The default presence penalty.
        """
        return self._presence_penalty

    @property
    def response_format(self) -> Optional[ResponseFormat]:
        """
        Get the default response format, if set.

        Returns:
            The default response format.
        """
        return self._response_format

    @property
    def seed(self) -> Optional[int]:
        """
        Get the default seed value, if set.

        Returns:
            The default seed value.
        """
        return self._seed

    @property
    def stop(self) -> Union[Optional[str], List[str]]:
        """
        Get the default stop value, if set.

        Returns:
            The default stop value.
        """
        return self._stop

    @property
    def stream(self) -> Optional[bool]:
        """
         Get the default stream bool, if set.

         Returns:
             The default stream bool.
         """
        return self._stream

    @property
    def stream_options(self) -> Optional[ChatCompletionStreamOptionsParam]:
        """
         Get the default stream options, if set.

         Returns:
             The default stream options.
         """
        return self._stream_options

    @property
    def temperature(self) -> Optional[float]:
        """
         Get the default temperature, if set.

         Returns:
             The default temperature.
         """
        return self._temperature

    @property
    def top_p(self) -> Optional[float]:
        """
         Get the default top p value, if set.

         Returns:
             The default top p value.
         """
        return self._top_p

    @property
    def tools(self) -> Optional[Iterable[ChatCompletionToolParam]]:
        """
         Get the default tools, if set.

         Returns:
             The default tools.
         """
        return self._tools

    @property
    def tool_choice(self) -> Optional[ChatCompletionToolChoiceOptionParam]:
        """
         Get the default tool choice, if set.

         Returns:
             The default tool choice.
         """
        return self._tool_choice

    @property
    def parallel_tool_calls(self) -> Optional[bool]:
        """
         Get the default parallel tool calls bool, if set.

         Returns:
             The default parallel tool calls bool.
         """
        return self._parallel_tool_calls

    @property
    def use_custom_keys(self) -> bool:
        """
         Get the default use custom keys bool, if set.

         Returns:
             The default use custom keys bool.
         """
        return self._use_custom_keys

    @property
    def tags(self) -> Optional[List[str]]:
        """
         Get the default tags, if set.

         Returns:
             The default tags.
         """
        return self._tags

    @property
    def drop_params(self) -> Optional[bool]:
        """
         Get the default drop_params bool, if set.

         Returns:
             The default drop_params bool.
         """
        return self._drop_params

    @property
    def region(self) -> Optional[str]:
        """
         Get the default region, if set.

         Returns:
             The default region.
         """
        return self._region

    @property
    def log_query_body(self) -> Optional[bool]:
        """
         Get the default log query body bool, if set.

         Returns:
             The default log query body bool.
         """
        return self._log_query_body

    @property
    def log_response_body(self) -> Optional[bool]:
        """
         Get the default log response body bool, if set.

         Returns:
             The default log response body bool.
         """
        return self._log_response_body

    @property
    def return_full_completion(self) -> bool:
        """
          Get the default return full completion bool.

          Returns:
              The default return full completion bool.
          """
        return self._return_full_completion

    @property
    def cache(self) -> bool:
        """
          Get default the cache bool.

          Returns:
              The default cache bool.
          """
        return self._cache

    @property
    def extra_headers(self) -> Optional[Headers]:
        """
         Get the default extra headers, if set.

         Returns:
             The default extra headers.
         """
        return self._extra_headers

    @property
    def extra_query(self) -> Optional[Query]:
        """
         Get the default extra query, if set.

         Returns:
             The default extra query.
         """
        return self._extra_query

    @property
    def extra_body(self) -> Optional[Mapping[str, str]]:
        """
         Get the default extra body, if set.

         Returns:
             The default extra body.
         """
        return self._extra_body

    @property
    def default_prompt(self) -> Prompt:
        """
        Get the default prompt, if set.

        Returns:
              The default prompt.
        """
        return Prompt(
            **{f: getattr(self, f) for f in Prompt.model_fields if
               hasattr(self, f) and getattr(self, f)}
        )

    # Setters #
    # --------#

    def set_system_message(self, value: str) -> Self:
        """
        Set the default system message.  # noqa: DAR101.

        Args:
            value: The default system message.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._system_message = value
        return self

    def set_messages(
            self,
            value: Union[List[ChatCompletionMessageParam],
                         Dict[str, List[ChatCompletionMessageParam]]]
    ) -> Self:
        """
        Set the default messages.  # noqa: DAR101.

        Args:
            value: The default messages.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._messages = value
        return self

    def set_frequency_penalty(self, value: float) -> Self:
        """
        Set the default frequency penalty.  # noqa: DAR101.

        Args:
            value: The default frequency penalty.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._frequency_penalty = value
        return self

    def set_logit_bias(self, value: Dict[str, int]) -> Self:
        """
        Set the default logit bias.  # noqa: DAR101.

        Args:
            value: The default logit bias.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._logit_bias = value
        return self

    def set_logprobs(self, value: bool) -> Self:
        """
        Set the default logprobs.  # noqa: DAR101.

        Args:
            value: The default logprobs.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._logprobs = value
        return self

    def set_top_logprobs(self, value: int) -> Self:
        """
        Set the default top logprobs.  # noqa: DAR101.

        Args:
            value: The default top logprobs.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._top_logprobs = value
        return self

    def set_max_completion_tokens(self, value: int) -> Self:
        """
        Set the default max tokens.  # noqa: DAR101.

        Args:
            value: The default max tokens.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._max_completion_tokens = value
        return self

    def set_n(self, value: int) -> Self:
        """
        Set the default n value.  # noqa: DAR101.

        Args:
            value: The default n value.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._n = value
        return self

    def set_presence_penalty(self, value: float) -> Self:
        """
        Set the default presence penalty.  # noqa: DAR101.

        Args:
            value: The default presence penalty.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._presence_penalty = value
        return self

    def set_response_format(self, value: ResponseFormat) -> Self:
        """
        Set the default response format.  # noqa: DAR101.

        Args:
            value: The default response format.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._response_format = value
        return self

    def set_seed(self, value: int) -> Self:
        """
        Set the default seed value.  # noqa: DAR101.

        Args:
            value: The default seed value.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._seed = value
        return self

    def set_stop(self, value: Union[str, List[str]]) -> Self:
        """
        Set the default stop value.  # noqa: DAR101.

        Args:
            value: The default stop value.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._stop = value
        return self

    def set_stream(self, value: bool) -> Self:
        """
        Set the default stream bool.  # noqa: DAR101.

        Args:
            value: The default stream bool.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._stream = value
        return self

    def set_stream_options(self, value: ChatCompletionStreamOptionsParam) -> Self:
        """
        Set the default stream options.  # noqa: DAR101.

        Args:
            value: The default stream options.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._stream_options = value
        return self

    def set_temperature(self, value: float) -> Self:
        """
        Set the default temperature.  # noqa: DAR101.

        Args:
            value: The default temperature.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._temperature = value
        return self

    def set_top_p(self, value: float) -> Self:
        """
        Set the default top p value.  # noqa: DAR101.

        Args:
            value: The default top p value.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._top_p = value
        return self

    def set_tools(self, value: Iterable[ChatCompletionToolParam]) -> Self:
        """
        Set the default tools.  # noqa: DAR101.

        Args:
            value: The default tools.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._tools = value
        return self

    def set_tool_choice(self, value: ChatCompletionToolChoiceOptionParam) -> Self:
        """
        Set the default tool choice.  # noqa: DAR101.

        Args:
            value: The default tool choice.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._tool_choice = value
        return self

    def set_parallel_tool_calls(self, value: bool) -> Self:
        """
        Set the default parallel tool calls bool.  # noqa: DAR101.

        Args:
            value: The default parallel tool calls bool.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._parallel_tool_calls = value
        return self

    def set_use_custom_keys(self, value: bool) -> Self:
        """
        Set the default use custom keys bool.  # noqa: DAR101.

        Args:
            value: The default use custom keys bool.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._use_custom_keys = value
        return self

    def set_tags(self, value: List[str]) -> Self:
        """
        Set the default tags.  # noqa: DAR101.

        Args:
            value: The default tags.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._tags = value
        return self

    def set_drop_params(self, value: bool) -> Self:
        """
        Set the default drop params bool.  # noqa: DAR101.

        Args:
            value: The default drop params bool.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._drop_params = value
        return self

    def set_region(self, value: str) -> Self:
        """
        Set the default region.  # noqa: DAR101.

        Args:
            value: The default region.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._region = value
        return self

    def set_log_query_body(self, value: bool) -> Self:
        """
        Set the default log query body bool.  # noqa: DAR101.

        Args:
            value: The default log query body bool.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._log_query_body = value
        return self

    def set_log_response_body(self, value: bool) -> Self:
        """
        Set the default log response body bool.  # noqa: DAR101.

        Args:
            value: The default log response body bool.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._log_response_body = value
        return self

    def set_return_full_completion(self, value: bool) -> Self:
        """
        Set the default return full completion bool.  # noqa: DAR101.

        Args:
            value: The default return full completion bool.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._return_full_completion = value
        return self

    def set_cache(self, value: bool) -> Self:
        """
        Set the default cache bool.  # noqa: DAR101.

        Args:
            value: The default cache bool.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._cache = value
        return self

    def set_extra_headers(self, value: Headers) -> Self:
        """
        Set the default extra headers.  # noqa: DAR101.

        Args:
            value: The default extra headers.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._extra_headers = value
        return self

    def set_extra_query(self, value: Query) -> Self:
        """
        Set the default extra query.  # noqa: DAR101.

        Args:
            value: The default extra query.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._extra_query = value
        return self

    def set_extra_body(self, value: Body) -> Self:
        """
        Set the default extra body.  # noqa: DAR101.

        Args:
            value: The default extra body.

        Returns:
            This client, useful for chaining inplace calls.
        """
        self._extra_body = value
        return self

    def set_default_prompt(self, value: Prompt) -> Self:
        """
        Set the default prompt.  # noqa: DAR101.

        Args:
              The default prompt.

        Returns:
            This client, useful for chaining inplace calls.
        """
        for f in value.model_fields:
            if hasattr(self, f):
                getattr(self, "set_" + f)(getattr(value, f))
        return self

    # Generate #
    # ---------#

    def generate(
            self,
            user_message: Optional[str] = None,
            system_message: Optional[str] = None,
            messages: Optional[
                Union[List[ChatCompletionMessageParam],
                      Dict[str, List[ChatCompletionMessageParam]]]] = None,
            *,
            frequency_penalty: Optional[float] = None,
            logit_bias: Optional[Dict[str, int]] = None,
            logprobs: Optional[bool] = None,
            top_logprobs: Optional[int] = None,
            max_completion_tokens: Optional[int] = None,
            n: Optional[int] = None,
            presence_penalty: Optional[float] = None,
            response_format: Optional[ResponseFormat] = None,
            seed: Optional[int] = None,
            stop: Union[Optional[str], List[str]] = None,
            stream: Optional[bool] = None,
            stream_options: Optional[ChatCompletionStreamOptionsParam] = None,
            temperature: Optional[float] = None,
            top_p: Optional[float] = None,
            tools: Optional[Iterable[ChatCompletionToolParam]] = None,
            tool_choice: Optional[ChatCompletionToolChoiceOptionParam] = None,
            parallel_tool_calls: Optional[bool] = None,
            # platform arguments
            use_custom_keys: Optional[bool] = None,
            tags: Optional[List[str]] = None,
            drop_params: Optional[bool] = None,
            region: Optional[str] = None,
            log_query_body: Optional[bool] = None,
            log_response_body: Optional[bool] = None,
            # python client arguments
            return_full_completion: Optional[bool] = None,
            cache: Optional[bool] = None,
            # passthrough arguments
            extra_headers: Optional[Headers] = None,
            extra_query: Optional[Query] = None,
            **kwargs,
    ):
        """Generate a ChatCompletion response for the specified endpoint,
        from the provided query parameters.

        Args:
            user_message: A string containing the user message.
            If provided, messages must be None.

            system_message: An optional string containing the system message. This
            always appears at the beginning of the list of messages.

            messages: A list of messages comprising the conversation so far, or
            optionally a dictionary of such messages, with clients as the keys in the
            case of multi-llm clients. This will be appended to the system_message if it
            is not None, and any user_message will be appended if it is not None.

            frequency_penalty: Number between -2.0 and 2.0. Positive values penalize new
            tokens based on their existing frequency in the text so far, decreasing the
            model's likelihood to repeat the same line verbatim.

            logit_bias: Modify the likelihood of specified tokens appearing in the
            completion. Accepts a JSON object that maps tokens (specified by their token
            ID in the tokenizer) to an associated bias value from -100 to 100.
            Mathematically, the bias is added to the logits generated by the model prior
            to sampling. The exact effect will vary per model, but values between -1 and
            1 should decrease or increase likelihood of selection; values like -100 or
            100 should result in a ban or exclusive selection of the relevant token.

            logprobs: Whether to return log probabilities of the output tokens or not.
            If true, returns the log probabilities of each output token returned in the
            content of message.

            top_logprobs: An integer between 0 and 20 specifying the number of most
            likely tokens to return at each token position, each with an associated log
            probability. logprobs must be set to true if this parameter is used.

            max_completion_tokens: The maximum number of tokens that can be generated in
            the chat completion. The total length of input tokens and generated tokens
            is limited by the model's context length. Defaults value is 1024. Uses the
            provider's default max_completion_tokens when None is explicitly passed.

            n: How many chat completion choices to generate for each input message. Note
            that you will be charged based on the number of generated tokens across all
            of the choices. Keep n as 1 to minimize costs.

            presence_penalty: Number between -2.0 and 2.0. Positive values penalize new
            tokens based on whether they appear in the text so far, increasing the
            model's likelihood to talk about new topics.

            response_format: An object specifying the format that the model must output.
            Setting to `{ "type": "json_schema", "json_schema": {...} }` enables
            Structured Outputs which ensures the model will match your supplied JSON
            schema. Learn more in the Structured Outputs guide. Setting to
            `{ "type": "json_object" }` enables JSON mode, which ensures the message the
            model generates is valid JSON.

            seed: If specified, a best effort attempt is made to sample
            deterministically, such that repeated requests with the same seed and
            parameters should return the same result. Determinism is not guaranteed, and
            you should refer to the system_fingerprint response parameter to monitor
            changes in the backend.

            stop: Up to 4 sequences where the API will stop generating further tokens.

            stream: If True, generates content as a stream. If False, generates content
            as a single response. Defaults to False.

            stream_options: Options for streaming response. Only set this when you set
            stream: true.

            temperature:  What sampling temperature to use, between 0 and 2.
            Higher values like 0.8 will make the output more random,
            while lower values like 0.2 will make it more focused and deterministic.
            It is generally recommended to alter this or top_p, but not both.
            Default value is 1.0. Defaults to the provider's default temperature when
            None is explicitly passed.

            top_p: An alternative to sampling with temperature, called nucleus sampling,
            where the model considers the results of the tokens with top_p probability
            mass. So 0.1 means only the tokens comprising the top 10% probability mass
            are considered. Generally recommended to alter this or temperature, but not
            both.

            tools: A list of tools the model may call. Currently, only functions are
            supported as a tool. Use this to provide a list of functions the model may
            generate JSON inputs for. A max of 128 functions are supported.

            tool_choice: Controls which (if any) tool is called by the
            model. none means the model will not call any tool and instead generates a
            message. auto means the model can pick between generating a message or
            calling one or more tools. required means the model must call one or more
            tools. Specifying a particular tool via
            `{ "type": "function", "function": {"name": "my_function"} }`
            forces the model to call that tool.
            none is the default when no tools are present. auto is the default if tools
            are present.

            parallel_tool_calls: Whether to enable parallel function calling during tool
            use.

            use_custom_keys:  Whether to use custom API keys or our unified API keys
            with the backend provider. Defaults to False.

            tags: Arbitrary number of tags to classify this API query as needed. Helpful
            for generally grouping queries across tasks and users, for logging purposes.

            drop_params: Whether or not to drop unsupported OpenAI params by the
            provider you’re using.

            region: A string used to represent the region where the endpoint is
            accessed. Only relevant for on-prem deployments with certain providers like
            `vertex-ai`, `aws-bedrock` and `azure-ml`, where the endpoint is being
            accessed through a specified region.

            log_query_body: Whether to log the contents of the query json body.

            log_response_body: Whether to log the contents of the response json body.

            return_full_completion: If False, only return the message content
            chat_completion.choices[0].message.content.strip(" ") from the OpenAI
            return. Otherwise, the full response chat_completion is returned.
            Defaults to False.

            cache: If True, then the arguments will be stored in a local cache file, and
            any future calls with identical arguments will read from the cache instead
            of running the LLM query. This can help to save costs and also debug
            multi-step LLM applications, while keeping early steps fixed.
            This argument only has any effect when stream=False. Defaults to False.

            extra_headers: Additional "passthrough" headers for the request which are
            provider-specific, and are not part of the OpenAI standard. They are handled
            by the provider-specific API.

            extra_query: Additional "passthrough" query parameters for the request which
            are provider-specific, and are not part of the OpenAI standard. They are
            handled by the provider-specific API.

            kwargs: Additional "passthrough" JSON properties for the body of the
            request, which are provider-specific, and are not part of the OpenAI
            standard. They will be handled by the provider-specific API.

        Returns:
            If stream is True, returns a generator yielding chunks of content.
            If stream is False, returns a single string response.

        Raises:
            UnifyError: If an error occurs during content generation.
        """
        return self._generate(
            user_message,
            _default(system_message, self._system_message),
            _default(messages, self._messages),
            frequency_penalty=_default(frequency_penalty, self._frequency_penalty),
            logit_bias=_default(logit_bias, self._logit_bias),
            logprobs=_default(logprobs, self._logprobs),
            top_logprobs=_default(top_logprobs, self._top_logprobs),
            max_completion_tokens=_default(max_completion_tokens,
                                           self._max_completion_tokens),
            n=_default(n, self._n),
            presence_penalty=_default(presence_penalty, self._presence_penalty),
            response_format=_default(response_format, self._response_format),
            seed=_default(seed, self._seed),
            stop=_default(stop, self._stop),
            stream=_default(stream, self._stream),
            stream_options=_default(stream_options, self._stream_options),
            temperature=_default(temperature, self._temperature),
            top_p=_default(top_p, self._top_p),
            tools=_default(tools, self._tools),
            tool_choice=_default(tool_choice, self._tool_choice),
            parallel_tool_calls=_default(parallel_tool_calls,
                                         self._parallel_tool_calls),
            # platform arguments
            use_custom_keys=_default(use_custom_keys, self._use_custom_keys),
            tags=_default(tags, self._tags),
            drop_params=_default(drop_params, self._drop_params),
            region=_default(region, self._region),
            log_query_body=_default(log_query_body, self._log_query_body),
            log_response_body=_default(log_response_body, self._log_response_body),
            # python client arguments
            return_full_completion=True if _default(tools, self._tools) else
            _default(return_full_completion, self._return_full_completion),
            cache=_default(cache, self._cache),
            # passthrough arguments
            extra_headers=_default(extra_headers, self._extra_headers),
            extra_query=_default(extra_query, self._extra_query),
            **{**self._extra_body, **kwargs},
        )

    # Credits #
    # --------#

    def get_credit_balance(self) -> Union[float, None]:
        # noqa: DAR201, DAR401
        """
        Get the remaining credits left on your account.

        Returns:
            The remaining credits on the account if successful, otherwise None.
        Raises:
            BadRequestError: If there was an HTTP error.
            ValueError: If there was an error parsing the JSON response.
        """
        url = f"{BASE_URL}/credits"
        headers = {
            "accept": "application/json",
            "Authorization": f"Bearer {self._api_key}",
        }
        try:
            response = requests.get(url, headers=headers, timeout=10)
            response.raise_for_status()
            return response.json()["credits"]
        except requests.RequestException as e:
            raise requests.RequestException(
                "There was an error with the request.") from e
        except (KeyError, ValueError) as e:
            raise ValueError("Error parsing JSON response.") from e

    # Abstract Methods #
    # -----------------#

    @abstractmethod
    def _generate(
            self,
            user_message: str,
            system_message: str,
            messages: Optional[
                Union[List[ChatCompletionMessageParam],
                      Dict[str, List[ChatCompletionMessageParam]]]],
            *,
            frequency_penalty: Optional[float],
            logit_bias: Optional[Dict[str, int]],
            logprobs: Optional[bool],
            top_logprobs: Optional[int],
            max_completion_tokens: Optional[int],
            n: Optional[int],
            presence_penalty: Optional[float],
            response_format: Optional[ResponseFormat],
            seed: Optional[int],
            stop: Union[Optional[str], List[str]],
            stream: Optional[bool],
            stream_options: Optional[ChatCompletionStreamOptionsParam],
            temperature: Optional[float],
            top_p: Optional[float],
            tools: Optional[Iterable[ChatCompletionToolParam]],
            tool_choice: Optional[ChatCompletionToolChoiceOptionParam],
            parallel_tool_calls: Optional[bool],
            # platform arguments
            use_custom_keys: bool,
            tags: Optional[List[str]],
            drop_params: Optional[bool],
            region: Optional[str],
            log_query_body: Optional[bool],
            log_response_body: Optional[bool],
            # python client arguments
            return_full_completion: bool,
            cache: bool,
            # passthrough arguments
            extra_headers: Optional[Headers],
            extra_query: Optional[Query],
            **kwargs,
    ):
        raise NotImplementedError



def create_router_config():
    raise NotImplementedError


def get_router_config():
    raise NotImplementedError


def delete_router_config():
    raise NotImplementedError


def rename_router_config():
    raise NotImplementedError


def list_router_configs():
    raise NotImplementedError



import abc
from abc import abstractmethod
from typing import Type, Union

from unify.agent import Agent
from unify.chat.clients import _Client
from unify.evaluation import Evaluation
from unify.casting import cast
from unify.types import Score, Prompt, ChatCompletion


class Evaluator(abc.ABC):

    @property
    @abstractmethod
    def class_config(self) -> Type[Score]:
        raise NotImplemented

    @abstractmethod
    def _evaluate(
            self,
            prompt: Union[str, Prompt],
            response: Union[str, ChatCompletion],
            **kwargs
    ) -> Union[bool, float, Score]:
        """
        Evaluate the given response for this input prompt, with optional extra data.

        Args:
            prompt: The user message or the full input prompt being responded to.

            response: The response which is being evaluated, either as just the most
            recent assistant message, or the full chat completion.

            kwargs: Extra information relevant to the prompt, as is stored in the Datum.

        Returns:
            The score, either as a boolean, a float, or the full Score instance.
        """
        raise NotImplemented

    def evaluate(
            self,
            prompt: Union[str, Prompt],
            response: Union[ChatCompletion, str],
            agent: Union[str, _Client, Agent],
            **kwargs
    ):
        """
        Evaluate the given response for this input prompt, with optional extra data.

        Args:
            prompt: The user message or the full input prompt being responded to.

            response: The response which is being evaluated, either as just the most
            recent assistant message, or the full chat completion.

            agent: The agent that made the response, which is being evaluated.

            kwargs: Extra information relevant to the prompt, as is stored in the Datum.

        Returns:
            An Evaluation instance, containing the prompt, response, agent, score and
            optional extra data used during the evaluation.
        """

        # upcast or downcast inputs

        # get type hints for self._evaluation, if they exist
        eval_ann = self._evaluate.__annotations__
        # upcast or downcast prompt to the expected type
        expected_prompt_type = eval_ann["prompt"] if "prompt" in eval_ann else Prompt
        prompt = cast(prompt, expected_prompt_type)
        # upcast or downcast response to the expected type
        expected_response_type = eval_ann["response"] if "response" in eval_ann \
            else ChatCompletion
        response = cast(response, expected_response_type)

        # perform the evaluation
        score = self._evaluate(prompt, response, **kwargs)

        # upcast to full type for storage in Evaluation

        # prompt upcasting
        prompt = cast(prompt, Prompt)
        # response upcasting
        response = cast(response, ChatCompletion)
        # score upcasting
        score = cast(score, self.class_config)
        # return evaluation
        return Evaluation(
            prompt=prompt,
            response=response,
            agent=agent,
            score=score,
            **kwargs
        )



import json
import requests
from typing import List, Dict, Optional, Union

from unify import BASE_URL
from unify.types import Datum
from .helpers import _validate_api_key, _res_to_list


def _upload_dataset_from_str(
    name: str, content: str, api_key: Optional[str] = None
) -> str:
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    files = {"file": ("dataset", content, "application/x-jsonlines")}
    data = {"name": name}
    # Send POST request to the /dataset endpoint
    response = requests.post(
        BASE_URL + "/dataset", headers=headers, data=data, files=files
    )
    response.raise_for_status()
    return json.loads(response.text)["info"]


def upload_dataset_from_file(
    name: str, path: str, api_key: Optional[str] = None
) -> str:
    """
    Uploads a local file as a dataset to the platform.

    Args:
        name: Name given to the uploaded dataset.
        path: Path to the file to be uploaded.
        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        Info msg with the response from the HTTP endpoint.
    Raises:
        ValueError: If there was an HTTP error.
    """
    with open(path, "rb") as f:
        content_str = f.read().decode()
    return _upload_dataset_from_str(name, content_str, api_key)


def upload_dataset_from_dictionary(
    name: str, content: List[Dict[str, str]], api_key: Optional[str] = None
) -> str:
    """
    Uploads a list of dictionaries as a dataset to the platform.
    Each dictionary in the list must contain a `prompt` key.

    Args:
        name: Name given to the uploaded dataset.
        content: Path to the file to be uploaded.
        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        Info msg with the response from the HTTP endpoint.
    Raises:
        ValueError: If there was an HTTP error.
    """
    content_str = "\n".join([json.dumps(d) for d in content])
    return _upload_dataset_from_str(name, content_str, api_key)


def download_dataset(
    name: str,
    path: Optional[str] = None,
    raw_return: bool = False,
    api_key: Optional[str] = None,
) -> Union[List[Datum], None]:
    """
    Downloads a dataset from the platform.

    Args:
        name: Name of the dataset to download.
        path: If specified, path to save the dataset.
        raw_return: Whether to provide the raw return, with extra meta-data.
        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        If path is not specified, returns the dataset content, if specified, returns
        None.
    Raises:
        ValueError: If there was an HTTP error.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    params = {"name": name}
    # Send GET request to the /dataset endpoint
    response = requests.get(BASE_URL + "/dataset", headers=headers, params=params)
    response.raise_for_status()
    if path:
        with open(path, "w+") as f:
            f.write("\n".join([json.dumps(d) for d in json.loads(response.text)]))
            return None
    ret = json.loads(response.text)
    if raw_return:
        return ret
    return [Datum(
        **{k: v for k, v in item.items() if k not in ("id", "num_tokens", "timestamp")}
    ) for item in ret]


def delete_dataset(name: str, api_key: Optional[str] = None) -> str:
    """
    Deletes a dataset from the platform.

    Args:
        name: Name given to the uploaded dataset.
        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        str: Info msg with the response from the HTTP endpoint.
    Raises:
        ValueError: If there was an HTTP error.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    params = {"name": name}
    # Send DELETE request to the /dataset endpoint
    response = requests.delete(BASE_URL + "/dataset", headers=headers, params=params)
    response.raise_for_status()
    return json.loads(response.text)["info"]


def rename_dataset(name: str, new_name: str, api_key: Optional[str] = None):
    """
    Renames a dataset.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    params = {"name": name, "new_name": new_name}
    response = requests.post(
        BASE_URL + "/dataset/rename", headers=headers, params=params
    )
    response.raise_for_status()
    return response.json()


def list_datasets(api_key: Optional[str] = None) -> List[str]:
    """
    Fetches a list of all uploaded datasets.

    Args:
        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        List with the names of the uploaded datasets.
    Raises:
        ValueError: If there was an HTTP error.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    # Send GET request to the /dataset/list endpoint
    response = requests.get(BASE_URL + "/dataset/list", headers=headers)
    response.raise_for_status()
    return _res_to_list(response)


def add_data(
        name: str,
        data: Union[Dict, List[Dict]],
        api_key: Optional[str] = None
):
    """
    Adds data to a dataset.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    body = {"name": name, "data": data}
    response = requests.post(
        BASE_URL + "/dataset/data", headers=headers, json=body
    )
    response.raise_for_status()
    return response.json()


def delete_data(
        name: str,
        data: Union[int, List[int], Dict, List[Dict]],
        api_key: Optional[str] = None
):
    """
    Delete data from a dataset, either by id or by value
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    if isinstance(data, list) and not data:
        return {"info": "data argument was empty. Nothing to delete."}
    if isinstance(data, dict) or (isinstance(data, list) and isinstance(data[0], dict)):
        # ToDo: remove this logic once delete-by-value is implemented in the REST API
        upstream_data = download_dataset(name, raw_return=True, api_key=api_key)
        upstream_data_pruned = [
            {k: v for k, v in item.items()
             if k not in ("id", "num_tokens", "timestamp")}
            for item in upstream_data
        ]
        data_ids = [
            d["id"] for d, dp in zip(upstream_data, upstream_data_pruned) if dp in data
        ]
        # ToDo end
    else:
        data_ids = data
    params = {"name": name, "data_ids": data_ids}
    response = requests.delete(
        BASE_URL + "/dataset/data", headers=headers, params=params
    )
    response.raise_for_status()
    return response.json()



import abc
import copy
import inspect
import rich.repr
from io import StringIO
from rich.console import Console
from pydantic import BaseModel, Extra, ConfigDict
from pydantic import create_model
from pydantic._internal._model_construction import ModelMetaclass
from typing import Optional, Union, Tuple, List, Dict, Mapping, Type
from openai._types import Query, Body
from openai.types.chat import (
    ChatCompletionToolParam,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionMessageParam,
    ChatCompletion as _ChatCompletion
)
from openai.types.completion_usage import CompletionUsage as _CompletionUsage
from openai.types.chat.completion_create_params import ResponseFormat
from openai.types.chat.chat_completion import \
    ChatCompletionMessage as _ChatCompletionMessage, Choice as _Choice

import unify

RICH_CONSOLE = Console(file=StringIO())


class _Formatted(abc.ABC):

    @staticmethod
    def _repr(to_print):
        # ToDO find more elegant way to do this
        global RICH_CONSOLE
        with RICH_CONSOLE.capture() as capture:
            RICH_CONSOLE.print(to_print)
        return capture.get()

    def __repr__(self) -> str:
        return self._repr(self)

    def __str__(self) -> str:
        return self._repr(self)


@rich.repr.auto
class _FormattedBaseModel(_Formatted, BaseModel):

    def _prune_dict(self, val, prune_policy):

        def keep(v, k=None, prune_pol=None):
            if v is None:
                return False
            if not prune_pol:
                return True
            if isinstance(prune_pol, dict) and \
                    "keep" not in prune_pol and "skip" not in prune_pol:
                return True
            if "keep" in prune_pol:
                if k not in prune_pol["keep"]:
                    return False
                prune_val = prune_pol["keep"][k]
                return prune_val is None or prune_val == v
            elif "skip" in prune_pol:
                if k not in prune_pol["skip"]:
                    return True
                prune_val = prune_pol["skip"][k]
                return prune_val is not None and prune_val != v
            else:
                raise Exception("expected prune_pol to contain either 'keep' or 'skip',"
                                "but neither were present: {}.".format(prune_pol))

        if not isinstance(val, dict) and not isinstance(val, list) and \
                not isinstance(val, tuple):
            return val
        elif isinstance(val, dict):
            return {
                k: self._prune_dict(
                    v, prune_policy[k] if
                    (isinstance(prune_policy, dict) and k in prune_policy) else None
                ) for k, v in val.items() if keep(v, k, prune_policy)
            }
        elif isinstance(val, list):
            return [
                self._prune_dict(
                    v, prune_policy[i] if
                    (isinstance(prune_policy, list ) and i < len(prune_policy))
                    else None
                ) for i, v in enumerate(val) if keep(v, prune_pol=prune_policy)
            ]
        else:
            return (
                self._prune_dict(
                    v, prune_policy[i] if
                    (isinstance(prune_policy, tuple) and i < len(prune_policy))
                    else None
                ) for i, v in enumerate(val) if keep(v, prune_pol=prune_policy)
            )

    def _prune_pydantic(self, val, dct):
        if not inspect.isclass(val) or not issubclass(val, BaseModel):
            return val
        config = {k: (self._prune_pydantic(val.model_fields[k].annotation, v),
                      val.model_fields[k].default) for k, v in dct.items()}
        if isinstance(val, ModelMetaclass):
            name = val.__qualname__
        else:
            name = val.__class__.__name__
        return create_model(name, **config)

    @staticmethod
    def _annotation(v):
        if hasattr(v, "annotation"):
            return v.annotation
        return type(v)

    @staticmethod
    def _default(v):
        if hasattr(v, "default"):
            return v.default
        return None

    def _prune(self):
        prune_policy = unify.key_repr(self)
        dct = self._prune_dict(self.dict(), prune_policy)
        fields = self.model_fields
        if self.model_extra is not None:
            fields = {**fields, **self.model_extra}
        config = {k: (self._prune_pydantic(self._annotation(fields[k]), v),
                      self._default(fields[k])) for k, v in dct.items()}
        return create_model(
            self.__class__.__name__,
            **config,
            __cls_kwargs__={"arbitrary_types_allowed": True}
        )(**dct)

    def __repr__(self) -> str:
        return self._repr(self._prune() if unify.repr_mode() == "concise" else self)

    def __str__(self) -> str:
        return self._repr(self._prune() if unify.repr_mode() == "concise" else self)

    def __rich_repr__(self):
        rep = self._prune() if unify.repr_mode() == "concise" else self
        for k in rep.model_fields:
            yield k, rep.__dict__[k]

    def full_repr(self):
        """
        Return the full un-pruned representation, regardless of the mode currently set.
        """
        with unify.ReprMode("verbose"):
            return self._repr(self)


class Prompt(_FormattedBaseModel):
    model_config = ConfigDict(extra="forbid")
    messages: Optional[List[ChatCompletionMessageParam]] = None
    frequency_penalty: Optional[float] = None
    logit_bias: Optional[Dict[str, int]] = None
    logprobs: Optional[bool] = None
    top_logprobs: Optional[int] = None
    max_completion_tokens: Optional[int] = None
    n: Optional[int] = None
    presence_penalty: Optional[float] = None
    response_format: Optional[ResponseFormat] = None
    seed: Optional[int] = None
    stop: Union[Optional[str], List[str]] = None
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    tools: Optional[List[ChatCompletionToolParam]] = None
    tool_choice: Optional[ChatCompletionToolChoiceOptionParam] = None
    parallel_tool_calls: Optional[bool] = None
    # extra_headers: Optional[Headers] = None  # ToDo: fix Omit error
    extra_headers: Optional[Mapping[str, str]] = None
    extra_query: Optional[Query] = None
    extra_body: Optional[Body] = None

    def __init__(
            self,
            user_message: Optional[str] = None,
            system_message: Optional[str] = None,
            **kwargs
    ):
        """
        Create Prompt instance.

        Args:
            user_message: The user message, optional.

            system_message: The system message, optional.

            kwargs: All fields expressed in the pydantic type.

        Returns:
            The pydantic Prompt instance.
        """
        if "messages" not in kwargs:
            kwargs["messages"] = list()
        if system_message:
            kwargs["messages"] = \
                [{"content": system_message, "role": "system"}] + \
                kwargs["messages"]
        if user_message:
            kwargs["messages"] += [{"content": user_message, "role": "user"}]
        if not kwargs["messages"]:
            kwargs["messages"] = None
        super().__init__(**kwargs)

    def __add__(self, other):
        return unify.Dataset(self) +\
               (other if isinstance(other, unify.Dataset) else unify.Dataset(other))

    def __sub__(self, other):
        return unify.Dataset(self) -\
               (other if isinstance(other, unify.Dataset) else unify.Dataset(other))

    def __radd__(self, other):
        return (other if isinstance(other, unify.Dataset) else unify.Dataset(other)) +\
               unify.Dataset(self)

    def __rsub__(self, other):
        return (other if isinstance(other, unify.Dataset) else unify.Dataset(other)) -\
               unify.Dataset(self)

    def __hash__(self):
        return hash(str(self))


class ChatCompletionMessage(_FormattedBaseModel, _ChatCompletionMessage):
    model_config = ConfigDict(extra="forbid")


class Choice(_FormattedBaseModel, _Choice):
    model_config = ConfigDict(extra="forbid")
    # only override pydantic types  which require FormattedBaseModel
    message: ChatCompletionMessage


class CompletionUsage(_FormattedBaseModel, _CompletionUsage):
    model_config = ConfigDict(extra="forbid")
    # cost is an extra field we've added, not in the OpenAI standard
    cost: float


class ChatCompletion(_FormattedBaseModel, _ChatCompletion):
    model_config = ConfigDict(extra="forbid")
    # only override pydantic types  which require FormattedBaseModel
    choices: List[Choice]
    usage: Optional[CompletionUsage] = None

    def __init__(self, assistant_message: Optional[str] = None, **kwargs):
        """
        Create ChatCompletion instance.

        Args:
            assistant_message: The assistant message, optional.

        Returns:
            The pydantic ChatCompletion instance.
        """
        if assistant_message:
            kwargs = {
                **dict(
                    id="",
                    choices=[
                        Choice(
                            finish_reason="stop",
                            index=0,
                            message=ChatCompletionMessage(
                                role="assistant",
                                content=assistant_message
                            )
                        )
                    ],
                    created=0,
                    model="",
                    object="chat.completion",
                    **kwargs
                ),
                **kwargs
            }
        super().__init__(**kwargs)


class Datum(_FormattedBaseModel, extra=Extra.allow):
    prompt: Prompt

    def __init__(self, prompt: Optional[Union[str, Prompt]]):
        """
        Create Datum instance.

        Args:
            prompt: The prompt, either as a user message or as the full prompt.

        Returns:
            The pydantic Datum instance.
        """
        if isinstance(prompt, str):
            prompt = Prompt(prompt)
        super().__init__(prompt=prompt)

    def __add__(self, other):
        return unify.Dataset(self) +\
               (other if isinstance(other, unify.Dataset) else unify.Dataset(other))

    def __sub__(self, other):
        return unify.Dataset(self) -\
               (other if isinstance(other, unify.Dataset) else unify.Dataset(other))

    def __radd__(self, other):
        return (other if isinstance(other, unify.Dataset) else unify.Dataset(other)) +\
               unify.Dataset(self)

    def __rsub__(self, other):
        return (other if isinstance(other, unify.Dataset) else unify.Dataset(other)) -\
               unify.Dataset(self)

    def __hash__(self):
        return hash(str(self))


class Score(_FormattedBaseModel, abc.ABC):
    model_config = ConfigDict(extra="forbid")
    score: Tuple[float, str]

    def __init__(self, value: float):
        """
        Create Score instance.

        Args:
            value: The value of the assigned score

        Returns:
            The pydantic Score instance, with associated value and class description
        """
        assert value in self.config,\
            "value {} passed is not a valid value, " \
            "based on the config for this Score class {}".format(value, self.config)
        super().__init__(score=(value, self.config[value]))

    @property
    @abc.abstractmethod
    def config(self) -> Dict[float, str]:
        raise NotImplementedError



import requests
from typing import Optional, Any, Dict

from unify import BASE_URL
from .helpers import _validate_api_key


def create_default_prompt(name: str, prompt: dict, api_key: Optional[str] = None):
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }

    url = f"{BASE_URL}/default_prompt"

    params = {
        "name": name,
        "prompt": prompt,
    }

    response = requests.post(url, headers=headers, json=params)
    response.raise_for_status()

    return response.json()


def get_default_prompt(name: str, api_key: Optional[str] = None):
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }

    url = f"{BASE_URL}/default_prompt"

    params = {
        "name": name,
    }

    response = requests.get(url, headers=headers, params=params)
    response.raise_for_status()

    return response.json()


def delete_default_prompt(name: str, api_key: Optional[str] = None):
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }

    url = f"{BASE_URL}/default_prompt"

    params = {
        "name": name,
    }

    response = requests.delete(url, headers=headers, params=params)
    response.raise_for_status()

    return response.json()


def rename_default_prompt(name: str, new_name: str, api_key: Optional[str] = None):
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }

    url = f"{BASE_URL}/default_prompt/rename"

    params = {
        "name": name,
        "new_name": new_name,
    }

    response = requests.post(url, headers=headers, params=params)
    response.raise_for_status()

    return response.json()


def list_default_prompts(api_key: Optional[str] = None):
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }

    url = f"{BASE_URL}/default_prompt/list"

    response = requests.get(url, headers=headers)
    response.raise_for_status()

    return response.json()



import requests
from typing import Optional

from unify import BASE_URL
from .helpers import _validate_api_key, _res_to_list


def get_credits(api_key: Optional[str] = None) -> float:
    """
    Returns the credits remaining in the user account, in USD.

    Args:
        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        The credits remaining in USD.
    Raises:
        ValueError: If there was an HTTP error.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    # Send GET request to the /get_credits endpoint
    response = requests.get(BASE_URL + "/credits", headers=headers)
    response.raise_for_status()
    return _res_to_list(response)["credits"]


def promo_code():
    raise NotImplementedError
